/*************************************************************************
* (C) 2011 AITNET ltd - Sofia/Bulgaria - <misho@aitbg.com>
*  by Michael Pounov <misho@openbsd-bg.org>
*
* $Author: misho $
* $Id: hooks.h,v 1.11 2015/07/02 22:45:00 misho Exp $
*
**************************************************************************
The ELWIX and AITNET software is distributed under the following
terms:

All of the documentation and software included in the ELWIX and AITNET
Releases is copyrighted by ELWIX - Sofia/Bulgaria <info@elwix.org>

Copyright 2004 - 2015
	by Michael Pounov <misho@elwix.org>.  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
This product includes software developed by Michael Pounov <misho@elwix.org>
ELWIX - Embedded LightWeight unIX and its contributors.
4. Neither the name of AITNET nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AITNET AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/
#ifndef __HOOKS_H
#define __HOOKS_H


/*
 * sched_hook_init() - Default INIT hook
 * @root = root task
 * @arg = unused
 * return: <0 errors and 0 ok
 */
void *sched_hook_init(void *root, void *arg __unused);

/*
 * sched_hook_fini() - Default FINI hook
 * @root = root task
 * @arg = unused
 * return: <0 errors and 0 ok
 */
void *sched_hook_fini(void *root, void *arg __unused);

/*
 * sched_hook_fetch() - Default FETCH hook
 * @root = root task
 * @arg = unused
 * return: NULL error or !=NULL fetched task
 */
void *sched_hook_fetch(void *root, void *arg __unused);

/*
 * sched_hook_cancel() - Default CANCEL hook
 * @task = current task
 * @arg = unused
 * return: <0 errors and 0 ok
 */
void *sched_hook_cancel(void *task, void *arg __unused);

/*
 * sched_hook_thread() - Default THREAD hook
 *
 * @task = current task
 * @arg = pthread attributes
 * return: <0 errors and 0 ok
 */
void *sched_hook_thread(void *task, void *arg);

/*
 * sched_hook_read() - Default READ hook
 * @task = current task
 * @arg = unused
 * return: <0 errors and 0 ok
 */
void *sched_hook_read(void *task, void *arg __unused);

/*
 * sched_hook_write() - Default WRITE hook
 * @task = current task
 * @arg = unused
 * return: <0 errors and 0 ok
 */
void *sched_hook_write(void *task, void *arg __unused);

/*
 * sched_hook_alarm() - Default ALARM hook
 *
 * @task = current task
 * @arg = unused
 * return: <0 errors and 0 ok
 */
void *sched_hook_alarm(void *task, void *arg __unused);

/*
 * sched_hook_node() - Default NODE hook
 *
 * @task = current task
 * @arg = unused
 * return: <0 errors and 0 ok
 */
void *sched_hook_node(void *task, void *arg __unused);

/*
 * sched_hook_proc() - Default PROC hook
 *
 * @task = current task
 * @arg = unused
 * return: <0 errors and 0 ok
 */
void *sched_hook_proc(void *task, void *arg __unused);

/*
 * sched_hook_signal() - Default SIGNAL hook
 *
 * @task = current task
 * @arg = unused
 * return: <0 errors and 0 ok
 */
void *sched_hook_signal(void *task, void *arg __unused);

/*
 * sched_hook_user() - Default USER hook
 *
 * @task = current task
 * @arg = unused
 * return: <0 errors and 0 ok
 */
#ifdef EVFILT_USER
void *sched_hook_user(void *task, void *arg __unused);
#endif

/*
 * sched_hook_exception() - Default EXCEPTION hook
 * @root = root task
 * @arg = custom handling: if arg == EV_EOF or other value; default: arg == NULL log errno
 * return: <0 errors and 0 ok
 */
void *sched_hook_exception(void *root, void *arg);

/*
 * sched_hook_condition() - Default CONDITION hook
 *
 * @root = root task
 * @arg = killState from schedRun()
 * return: NULL kill scheduler loop or !=NULL ok
 */
void *sched_hook_condition(void *root, void *arg);

/*
 * sched_hook_rtc() - Default RTC hook
 *
 * @task = current task
 * @arg = unused
 * return: <0 errors and 0 ok
 */
#if defined(HAVE_TIMER_CREATE) && defined(HAVE_TIMER_SETTIME)
void *sched_hook_rtc(void *task, void *arg __unused);
#endif


#endif
