/*************************************************************************
* (C) 2011 AITNET ltd - Sofia/Bulgaria - <misho@aitbg.com>
*  by Michael Pounov <misho@openbsd-bg.org>
*
* $Author: misho $
* $Id: aitsched.c,v 1.28 2015/07/22 19:50:45 misho Exp $
*
**************************************************************************
The ELWIX and AITNET software is distributed under the following
terms:

All of the documentation and software included in the ELWIX and AITNET
Releases is copyrighted by ELWIX - Sofia/Bulgaria <info@elwix.org>

Copyright 2004 - 2015
	by Michael Pounov <misho@elwix.org>.  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
This product includes software developed by Michael Pounov <misho@elwix.org>
ELWIX - Embedded LightWeight unIX and its contributors.
4. Neither the name of AITNET nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AITNET AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/
#include "global.h"
#include "hooks.h"


#pragma GCC visibility push(hidden)

int sched_Errno;
char sched_Error[STRSIZ];

#pragma GCC visibility pop


// sched_GetErrno() Get error code of last operation
int
sched_GetErrno()
{
	return sched_Errno;
}

// sched_GetError() Get error text of last operation
const char *
sched_GetError()
{
	return sched_Error;
}

// sched_SetErr() Set error to variables for internal use!!!
void
sched_SetErr(int eno, char *estr, ...)
{
	va_list lst;

	sched_Errno = eno;
	memset(sched_Error, 0, sizeof sched_Error);
	va_start(lst, estr);
	vsnprintf(sched_Error, sizeof sched_Error, estr, lst);
	va_end(lst);
}


/* string support functions directly imported from OpenBSD */

#ifndef HAVE_STRLCAT
/*
 * Appends src to string dst of size siz (unlike strncat, siz is the
 * full size of dst, not space left).  At most siz-1 characters
 * will be copied.  Always NUL terminates (unless siz <= strlen(dst)).
 * Returns strlen(src) + MIN(siz, strlen(initial dst)).
 * If retval >= siz, truncation occurred.
 */
size_t
strlcat(char * __restrict dst, const char * __restrict src, size_t siz)
{
	char *d = dst;
	const char *s = src;
	size_t n = siz;
	size_t dlen;

	/* Find the end of dst and adjust bytes left but don't go past end */
	while (n-- != 0 && *d != '\0')
		d++;
	dlen = d - dst;
	n = siz - dlen;

	if (n == 0)
		return(dlen + strlen(s));
	while (*s != '\0') {
		if (n != 1) {
			*d++ = *s;
			n--;
		}
		s++;
	}
	*d = '\0';

	return(dlen + (s - src));	/* count does not include NUL */
}
#endif
#ifndef HAVE_STRLCPY
/*
 * Copy src to string dst of size siz.  At most siz-1 characters
 * will be copied.  Always NUL terminates (unless siz == 0).
 * Returns strlen(src); if retval >= siz, truncation occurred.
 */
size_t
strlcpy(char * __restrict dst, const char * __restrict src, size_t siz)
{
	char *d = dst;
	const char *s = src;
	size_t n = siz;

	/* Copy as many bytes as will fit */
	if (n != 0) {
		while (--n != 0) {
			if ((*d++ = *s++) == '\0')
				break;
		}
	}

	/* Not enough room in dst, add NUL and traverse rest of src */
	if (n == 0) {
		if (siz != 0)
			*d = '\0';		/* NUL-terminate dst */
		while (*s++)
			;
	}

	return(s - src - 1);	/* count does not include NUL */
}
#endif


/* Init and prepare scheduler functions */

/*
 * schedRegisterHooks() - Register IO handles and bind tasks to it
 *
 * @root = root task
 * return: -1 error or 0 ok
 */
int
schedRegisterHooks(sched_root_task_t * __restrict root)
{
	assert(root);

	if (root->root_hooks.hook_root.fini)
		root->root_hooks.hook_root.fini(root, NULL);
	memset(&root->root_hooks, 0, sizeof root->root_hooks);

	root->root_hooks.hook_add.read = sched_hook_read;
	root->root_hooks.hook_add.write = sched_hook_write;
#if defined(HAVE_TIMER_CREATE) && defined(HAVE_TIMER_SETTIME) && defined(HAVE_TIMER_DELETE)
	root->root_hooks.hook_add.rtc = sched_hook_rtc;
#endif
#if SUP_ENABLE == KQ_SUPPORT
	root->root_hooks.hook_add.alarm = sched_hook_alarm;
	root->root_hooks.hook_add.node = sched_hook_node;
	root->root_hooks.hook_add.proc = sched_hook_proc;
	root->root_hooks.hook_add.signal = sched_hook_signal;
#ifdef EVFILT_USER
	root->root_hooks.hook_add.user = sched_hook_user;
#endif
#endif	/* KQ_SUPPORT */
#ifdef HAVE_LIBPTHREAD
	root->root_hooks.hook_add.thread = sched_hook_thread;
#endif

	root->root_hooks.hook_exec.cancel = sched_hook_cancel;
	root->root_hooks.hook_exec.fetch = sched_hook_fetch;
	root->root_hooks.hook_exec.exception = sched_hook_exception;

	root->root_hooks.hook_root.init = sched_hook_init;
	root->root_hooks.hook_root.fini = sched_hook_fini;
	return 0;
}

/*
 * schedInit() - Init scheduler
 *
 * @data = optional data if !=NULL
 * @datlen = data len if data is set
 * return: allocated root task if ok or NULL error
 */
sched_root_task_t *
schedInit(void ** __restrict data, size_t datlen)
{
	sched_root_task_t *root = NULL;
	int (*func)(sched_root_task_t *);
#ifdef HAVE_LIBPTHREAD
	register int i;
#endif

	root = malloc(sizeof(sched_root_task_t));
	if (!root) {
		LOGERR;
	} else {
		memset(root, 0, sizeof(sched_root_task_t));

		/* set default maximum regular task hit misses */
		root->root_miss = MAX_TASK_MISS;

		/* INFINIT polling period by default */
		sched_timespecinf(&root->root_poll);

#ifdef HAVE_LIBPTHREAD
		for (i = 0; i < taskMAX; i++)
			if ((errno = pthread_mutex_init(&root->root_mtx[i], NULL))) {
				LOGERR;
				while (i)
					pthread_mutex_destroy(&root->root_mtx[--i]);
				free(root);
				return NULL;
			}

		for (i = 0; i < taskMAX; i++)
			pthread_mutex_lock(&root->root_mtx[i]);
#endif

		TAILQ_INIT(&root->root_read);
		TAILQ_INIT(&root->root_write);
		TAILQ_INIT(&root->root_timer);
		TAILQ_INIT(&root->root_alarm);
		TAILQ_INIT(&root->root_rtc);
		TAILQ_INIT(&root->root_node);
		TAILQ_INIT(&root->root_proc);
		TAILQ_INIT(&root->root_signal);
		TAILQ_INIT(&root->root_aio);
		TAILQ_INIT(&root->root_lio);
		TAILQ_INIT(&root->root_user);
		TAILQ_INIT(&root->root_event);
		TAILQ_INIT(&root->root_task);
		TAILQ_INIT(&root->root_suspend);
		TAILQ_INIT(&root->root_ready);
		TAILQ_INIT(&root->root_unuse);
		TAILQ_INIT(&root->root_thread);

#ifdef HAVE_LIBPTHREAD
		for (i = 0; i < taskMAX; i++)
			pthread_mutex_unlock(&root->root_mtx[i]);
#endif

		if (data && *data) {
			if (datlen) {
				root->root_data.iov_base = *data;
				root->root_data.iov_len = datlen;
			} else { /* if datlen == 0, switch to callbacks init mode */
				 /* little hack :) for correct initialization of scheduler */
				func = (int(*)(sched_root_task_t*)) data;
				func(root);
			}
		}

		if (root->root_hooks.hook_root.init)
			root->root_hooks.hook_root.init(root, NULL);
	}

	return root;
}

/*
 * schedEnd() - End scheduler & free all resources
 *
 * @root = root task
 * return: -1 error or 0 ok
 */
int
schedEnd(sched_root_task_t ** __restrict root)
{
	sched_task_t *task, *tmp;
#ifdef HAVE_LIBPTHREAD
	register int i;
#endif

	if (!root || !*root)
		return -1;

#if 0
	TAILQ_FOREACH_SAFE(task, &(*root)->root_read, task_node, tmp)
		printf("read=%p\n", task);
	TAILQ_FOREACH_SAFE(task, &(*root)->root_write, task_node, tmp)
		printf("write=%p\n", task);
	TAILQ_FOREACH_SAFE(task, &(*root)->root_timer, task_node, tmp)
		printf("timer=%p\n", task);
	TAILQ_FOREACH_SAFE(task, &(*root)->root_alarm, task_node, tmp)
		printf("alarm=%p\n", task);
	TAILQ_FOREACH_SAFE(task, &(*root)->root_rtc, task_node, tmp)
		printf("rtc=%p\n", task);
	TAILQ_FOREACH_SAFE(task, &(*root)->root_node, task_node, tmp)
		printf("node=%p\n", task);
	TAILQ_FOREACH_SAFE(task, &(*root)->root_proc, task_node, tmp)
		printf("proc=%p\n", task);
	TAILQ_FOREACH_SAFE(task, &(*root)->root_signal, task_node, tmp)
		printf("signal=%p\n", task);
	TAILQ_FOREACH_SAFE(task, &(*root)->root_aio, task_node, tmp)
		printf("aio=%p\n", task);
	TAILQ_FOREACH_SAFE(task, &(*root)->root_lio, task_node, tmp)
		printf("lio=%p\n", task);
	TAILQ_FOREACH_SAFE(task, &(*root)->root_user, task_node, tmp)
		printf("user=%p\n", task);
	TAILQ_FOREACH_SAFE(task, &(*root)->root_event, task_node, tmp)
		printf("event=%p\n", task);
	TAILQ_FOREACH_SAFE(task, &(*root)->root_suspend, task_node, tmp)
		printf("suspend=%p\n", task);
	TAILQ_FOREACH_SAFE(task, &(*root)->root_ready, task_node, tmp)
		printf("ready=%p\n", task);
	TAILQ_FOREACH_SAFE(task, &(*root)->root_thread, task_node, tmp)
		printf("thread=%p\n", task);
	TAILQ_FOREACH_SAFE(task, &(*root)->root_task, task_node, tmp)
		printf("task=%p\n", task);
	TAILQ_FOREACH_SAFE(task, &(*root)->root_unuse, task_node, tmp)
		printf("unuse=%p\n", task);
	fflush(stdout);
#endif

	TAILQ_FOREACH_SAFE(task, &(*root)->root_read, task_node, tmp)
		schedCancel(task);
	TAILQ_FOREACH_SAFE(task, &(*root)->root_write, task_node, tmp)
		schedCancel(task);
	TAILQ_FOREACH_SAFE(task, &(*root)->root_timer, task_node, tmp)
		schedCancel(task);
	TAILQ_FOREACH_SAFE(task, &(*root)->root_alarm, task_node, tmp)
		schedCancel(task);
	TAILQ_FOREACH_SAFE(task, &(*root)->root_rtc, task_node, tmp)
		schedCancel(task);
	TAILQ_FOREACH_SAFE(task, &(*root)->root_node, task_node, tmp)
		schedCancel(task);
	TAILQ_FOREACH_SAFE(task, &(*root)->root_proc, task_node, tmp)
		schedCancel(task);
	TAILQ_FOREACH_SAFE(task, &(*root)->root_signal, task_node, tmp)
		schedCancel(task);
	TAILQ_FOREACH_SAFE(task, &(*root)->root_aio, task_node, tmp)
		schedCancel(task);
	TAILQ_FOREACH_SAFE(task, &(*root)->root_lio, task_node, tmp)
		schedCancel(task);
	TAILQ_FOREACH_SAFE(task, &(*root)->root_user, task_node, tmp)
		schedCancel(task);
	TAILQ_FOREACH_SAFE(task, &(*root)->root_event, task_node, tmp)
		schedCancel(task);
	TAILQ_FOREACH_SAFE(task, &(*root)->root_suspend, task_node, tmp)
		schedCancel(task);
	TAILQ_FOREACH_SAFE(task, &(*root)->root_ready, task_node, tmp)
		schedCancel(task);
	TAILQ_FOREACH_SAFE(task, &(*root)->root_thread, task_node, tmp)
		schedCancel(task);
	TAILQ_FOREACH_SAFE(task, &(*root)->root_task, task_node, tmp)
		schedCancel(task);

	SCHED_QLOCK((*root), taskUNUSE);
	TAILQ_FOREACH_SAFE(task, &(*root)->root_unuse, task_node, tmp) {
		TAILQ_REMOVE(&(*root)->root_unuse, task, task_node);
		free(task);
	}
	SCHED_QUNLOCK((*root), taskUNUSE);

	if ((*root)->root_hooks.hook_root.fini)
		(*root)->root_hooks.hook_root.fini(*root, NULL);

#ifdef HAVE_LIBPTHREAD
	for (i = 0; i < taskMAX; i++) {
		SCHED_QTRYLOCK(*root, i);
		SCHED_QUNLOCK(*root, i);
		pthread_mutex_destroy(&(*root)->root_mtx[i]);
	}
#endif

	free(*root);
	*root = NULL;
	return 0;
}

/*
 * schedCall() - Call task execution function
 *
 * @task = current task
 * return: !=NULL error or =NULL ok
 */
void *
schedCall(sched_task_t * __restrict task)
{
	void *ptr = (void*) -1;

	if (!task)
		return ptr;

	if (!TASK_ISLOCKED(task))
		TASK_LOCK(task);

	ptr = task->task_func(task);

	TASK_UNLOCK(task);
	return ptr;
}

/*
 * schedFetch() - Fetch ready task
 *
 * @root = root task
 * return: =NULL error or !=NULL ready task
 */
void *
schedFetch(sched_root_task_t * __restrict root)
{
	void *ptr;

	if (!root)
		return NULL;

	if (root->root_hooks.hook_exec.fetch)
		ptr = root->root_hooks.hook_exec.fetch(root, NULL);
	else
		ptr = NULL;

	return ptr;
}

/*
 * schedTrigger() - Triggering USER task
 *
 * @task = task
 * return: -1 error or 0 ok
 */
int
schedTrigger(sched_task_t * __restrict task)
{
#if SUP_ENABLE != KQ_SUPPORT
	sched_SetErr(ENOTSUP, "disabled kqueue support");
	return -1;
#else
#ifndef EVFILT_USER
	sched_SetErr(ENOTSUP, "Not supported kevent() filter");
	return -1;
#else
	struct kevent chg[1];
	struct timespec timeout = { 0, 0 };

	if (!task || !TASK_ROOT(task))
		return -1;

#ifdef __NetBSD__
	EV_SET(chg, TASK_VAL(task), EVFILT_USER, 0, NOTE_TRIGGER, 0, (intptr_t) TASK_VAL(task));
#else
	EV_SET(chg, TASK_VAL(task), EVFILT_USER, 0, NOTE_TRIGGER, 0, (void*) TASK_VAL(task));
#endif
	if (kevent(TASK_ROOT(task)->root_kq, chg, 1, NULL, 0, &timeout) == -1) {
		LOGERR;
		return -1;
	}

	return 0;
#endif
#endif	/* KQ_SUPPORT */
}

/*
 * schedQuery() - Query task in scheduler
 *
 * @task = task
 * return: -1 error, 0 found  and 1 not found
 */
int
schedQuery(sched_task_t * __restrict task)
{
	sched_queue_t *queue;
	sched_task_t *t;

	if (!task || !TASK_ROOT(task))
		return -1;	/* error */

	switch (TASK_TYPE(task)) {
		case taskREAD:
			queue = &TASK_ROOT(task)->root_read;
			break;
		case taskWRITE:
			queue = &TASK_ROOT(task)->root_write;
			break;
		case taskTIMER:
			queue = &TASK_ROOT(task)->root_timer;
			break;
		case taskALARM:
			queue = &TASK_ROOT(task)->root_alarm;
			break;
		case taskRTC:
			queue = &TASK_ROOT(task)->root_rtc;
			break;
		case taskNODE:
			queue = &TASK_ROOT(task)->root_node;
			break;
		case taskPROC:
			queue = &TASK_ROOT(task)->root_proc;
			break;
		case taskSIGNAL:
			queue = &TASK_ROOT(task)->root_signal;
			break;
		case taskAIO:
			queue = &TASK_ROOT(task)->root_aio;
			break;
		case taskLIO:
			queue = &TASK_ROOT(task)->root_lio;
			break;
		case taskUSER:
			queue = &TASK_ROOT(task)->root_user;
			break;
		case taskEVENT:
			queue = &TASK_ROOT(task)->root_event;
			break;
		case taskTASK:
			queue = &TASK_ROOT(task)->root_task;
			break;
		case taskSUSPEND:
			queue = &TASK_ROOT(task)->root_suspend;
			break;
		case taskREADY:
			queue = &TASK_ROOT(task)->root_ready;
			break;
		case taskTHREAD:
			queue = &TASK_ROOT(task)->root_thread;
			break;
		default:
			return 1;	/* not in queue */
	}
	if (queue)
		TAILQ_FOREACH(t, queue, task_node)
			if (TASK_ID(t) == TASK_ID(task))
				return 0;	/* found */

	return 1;	/* not in queue */
}

/*
 * schedQueryby() - Query task in scheduler by criteria
 *
 * @root = root task
 * @type = query from queue type, if =taskMAX query same task from all queues
 * @criteria = find task by criteria 
 * 	[ CRITERIA_ANY|CRITERIA_CALL|CRITERIA_ARG|CRITERIA_FD|CRITERIA_VAL|
 * 		CRITERIA_ID|CRITERIA_TS|CRITERIA_DATA|CRITERIA_DATLEN ]
 * @param = search parameter
 * return: -1 error, 0 found or 1 not found
 */
int
schedQueryby(sched_root_task_t * __restrict root, sched_task_type_t type, 
		u_char criteria, void *param)
{
	sched_task_t *task;
	sched_queue_t *queue;
	register int flg = 0;

	if (!root)
		return -1;
	/* if type == taskMAX check in all queues */
	if (type == taskMAX) {
		if ((flg = schedQueryby(root, taskREAD, criteria, param)) < 1)
			return flg;
		if ((flg = schedQueryby(root, taskWRITE, criteria, param)) < 1)
			return flg;
		if ((flg = schedQueryby(root, taskTIMER, criteria, param)) < 1)
			return flg;
		if ((flg = schedQueryby(root, taskALARM, criteria, param)) < 1)
			return flg;
		if ((flg = schedQueryby(root, taskRTC, criteria, param)) < 1)
			return flg;
		if ((flg = schedQueryby(root, taskNODE, criteria, param)) < 1)
			return flg;
		if ((flg = schedQueryby(root, taskPROC, criteria, param)) < 1)
			return flg;
		if ((flg = schedQueryby(root, taskSIGNAL, criteria, param)) < 1)
			return flg;
		if ((flg = schedQueryby(root, taskAIO, criteria, param)) < 1)
			return flg;
		if ((flg = schedQueryby(root, taskLIO, criteria, param)) < 1)
			return flg;
		if ((flg = schedQueryby(root, taskUSER, criteria, param)) < 1)
			return flg;
		if ((flg = schedQueryby(root, taskEVENT, criteria, param)) < 1)
			return flg;
		if ((flg = schedQueryby(root, taskTASK, criteria, param)) < 1)
			return flg;
		if ((flg = schedQueryby(root, taskSUSPEND, criteria, param)) < 1)
			return flg;
		if ((flg = schedQueryby(root, taskREADY, criteria, param)) < 1)
			return flg;
		if ((flg = schedQueryby(root, taskTHREAD, criteria, param)) < 1)
			return flg;
		return 1;	/* not found */
	}
	/* choosen queue */
	switch (type) {
		case taskREAD:
			queue = &root->root_read;
			break;
		case taskWRITE:
			queue = &root->root_write;
			break;
		case taskTIMER:
			queue = &root->root_timer;
			break;
		case taskALARM:
			queue = &root->root_alarm;
			break;
		case taskRTC:
			queue = &root->root_rtc;
			break;
		case taskNODE:
			queue = &root->root_node;
			break;
		case taskPROC:
			queue = &root->root_proc;
			break;
		case taskSIGNAL:
			queue = &root->root_signal;
			break;
		case taskAIO:
			queue = &root->root_aio;
			break;
		case taskLIO:
			queue = &root->root_lio;
			break;
		case taskUSER:
			queue = &root->root_user;
			break;
		case taskEVENT:
			queue = &root->root_event;
			break;
		case taskTASK:
			queue = &root->root_task;
			break;
		case taskSUSPEND:
			queue = &root->root_suspend;
			break;
		case taskREADY:
			queue = &root->root_ready;
			break;
		case taskTHREAD:
			queue = &root->root_thread;
			break;
		default:
			return 1;	/* not found */
	}

	TAILQ_FOREACH(task, queue, task_node) {
		switch (criteria) {
			case CRITERIA_ANY:
				return 0;		/* found */
			case CRITERIA_CALL:
				if (TASK_FUNC(task) == (sched_task_func_t) param)
					return 0;	/* found */
				break;
			case CRITERIA_ARG:
				if (TASK_ARG(task) == param)
					return 0;	/* found */
				break;
			case CRITERIA_FD:
				if (TASK_FD(task) == (intptr_t) param)
					return 0;	/* found */
				break;
			case CRITERIA_ID:
			case CRITERIA_VAL:
				if (TASK_VAL(task) == (u_long) param)
					return 0;	/* found */
				break;
			case CRITERIA_TS:
				if (!sched_timespeccmp(&TASK_TS(task), 
							(struct timespec*) param, -))
					return 0;	/* found */
				break;
			case CRITERIA_DATA:
				if (TASK_DATA(task) == param)
					return 0;	/* found */
				break;
			case CRITERIA_DATLEN:
				if (TASK_DATLEN(task) == (size_t) param)
					return 0;	/* found */
				break;
			default:
				sched_SetErr(EINVAL, "Invalid parameter criteria %d", 
						criteria);
				return 1;		/* not found */
		}
	}

	return 1;	/* not found */
}

/*
 * schedCancel() - Cancel task from scheduler
 *
 * @task = task
 * return: -1 error or 0 ok
 */
int
schedCancel(sched_task_t * __restrict task)
{
	sched_queue_t *queue;

	if (!task || !TASK_ROOT(task))
		return -1;

	if (TASK_ROOT(task)->root_hooks.hook_exec.cancel)
		if (TASK_ROOT(task)->root_hooks.hook_exec.cancel(task, NULL))
			return -1;

	switch (TASK_TYPE(task)) {
		case taskREAD:
			queue = &TASK_ROOT(task)->root_read;
			break;
		case taskWRITE:
			queue = &TASK_ROOT(task)->root_write;
			break;
		case taskTIMER:
			queue = &TASK_ROOT(task)->root_timer;
			break;
		case taskALARM:
			queue = &TASK_ROOT(task)->root_alarm;
			break;
		case taskRTC:
			queue = &TASK_ROOT(task)->root_rtc;
			break;
		case taskNODE:
			queue = &TASK_ROOT(task)->root_node;
			break;
		case taskPROC:
			queue = &TASK_ROOT(task)->root_proc;
			break;
		case taskSIGNAL:
			queue = &TASK_ROOT(task)->root_signal;
			break;
		case taskAIO:
			queue = &TASK_ROOT(task)->root_aio;
			break;
		case taskLIO:
			queue = &TASK_ROOT(task)->root_lio;
			break;
		case taskUSER:
			queue = &TASK_ROOT(task)->root_user;
			break;
		case taskEVENT:
			queue = &TASK_ROOT(task)->root_event;
			break;
		case taskTASK:
			queue = &TASK_ROOT(task)->root_task;
			break;
		case taskSUSPEND:
			queue = &TASK_ROOT(task)->root_suspend;
			break;
		case taskREADY:
			queue = &TASK_ROOT(task)->root_ready;
			break;
		case taskTHREAD:
			queue = &TASK_ROOT(task)->root_thread;
			break;
		default:
			queue = NULL;
	}
	if (queue)
		remove_task_from(task, queue);
	if (TASK_TYPE(task) != taskUNUSE)
		sched_unuseTask(task);

	return 0;
}

/*
 * schedCancelby() - Cancel task from scheduler by criteria
 *
 * @root = root task
 * @type = cancel from queue type, if =taskMAX cancel same task from all queues
 * @criteria = find task by criteria 
 * 	[ CRITERIA_ANY|CRITERIA_CALL|CRITERIA_ARG|CRITERIA_FD|CRITERIA_VAL|
 * 		CRITERIA_ID|CRITERIA_TS|CRITERIA_DATA|CRITERIA_DATLEN ]
 * @param = search parameter
 * @hook = custom cleanup hook function, may be NULL
 * return: -1 error, -2 error in sub-stage cancel execution, -3 error from custom hook or 0 ok
 */
int
schedCancelby(sched_root_task_t * __restrict root, sched_task_type_t type, 
		u_char criteria, void *param, sched_hook_func_t hook)
{
	sched_task_t *task, *tmp;
	sched_queue_t *queue;
	register int flg = 0;

	if (!root)
		return -1;
	/* if type == taskMAX check in all queues */
	if (type == taskMAX) {
		if (schedCancelby(root, taskREAD, criteria, param, hook))
			return -2;
		if (schedCancelby(root, taskWRITE, criteria, param, hook))
			return -2;
		if (schedCancelby(root, taskTIMER, criteria, param, hook))
			return -2;
		if (schedCancelby(root, taskALARM, criteria, param, hook))
			return -2;
		if (schedCancelby(root, taskRTC, criteria, param, hook))
			return -2;
		if (schedCancelby(root, taskNODE, criteria, param, hook))
			return -2;
		if (schedCancelby(root, taskPROC, criteria, param, hook))
			return -2;
		if (schedCancelby(root, taskSIGNAL, criteria, param, hook))
			return -2;
		if (schedCancelby(root, taskAIO, criteria, param, hook))
			return -2;
		if (schedCancelby(root, taskLIO, criteria, param, hook))
			return -2;
		if (schedCancelby(root, taskUSER, criteria, param, hook))
			return -2;
		if (schedCancelby(root, taskEVENT, criteria, param, hook))
			return -2;
		if (schedCancelby(root, taskTASK, criteria, param, hook))
			return -2;
		if (schedCancelby(root, taskSUSPEND, criteria, param, hook))
			return -2;
		if (schedCancelby(root, taskREADY, criteria, param, hook))
			return -2;
		if (schedCancelby(root, taskTHREAD, criteria, param, hook))
			return -2;
		return 0;
	}
	/* choosen queue */
	switch (type) {
		case taskREAD:
			queue = &root->root_read;
			break;
		case taskWRITE:
			queue = &root->root_write;
			break;
		case taskTIMER:
			queue = &root->root_timer;
			break;
		case taskALARM:
			queue = &root->root_alarm;
			break;
		case taskRTC:
			queue = &root->root_rtc;
			break;
		case taskNODE:
			queue = &root->root_node;
			break;
		case taskPROC:
			queue = &root->root_proc;
			break;
		case taskSIGNAL:
			queue = &root->root_signal;
			break;
		case taskAIO:
			queue = &root->root_aio;
			break;
		case taskLIO:
			queue = &root->root_lio;
			break;
		case taskUSER:
			queue = &root->root_user;
			break;
		case taskEVENT:
			queue = &root->root_event;
			break;
		case taskTASK:
			queue = &root->root_task;
			break;
		case taskSUSPEND:
			queue = &root->root_suspend;
			break;
		case taskREADY:
			queue = &root->root_ready;
			break;
		case taskTHREAD:
			queue = &root->root_thread;
			break;
		default:
			return 0;
	}

	SCHED_QLOCK(root, type);
	TAILQ_FOREACH_SAFE(task, queue, task_node, tmp) {
		flg ^= flg;
		switch (criteria) {
			case CRITERIA_ANY:
				flg = 1;
				break;
			case CRITERIA_CALL:
				if (TASK_FUNC(task) == (sched_task_func_t) param)
					flg = 1;
				break;
			case CRITERIA_ARG:
				if (TASK_ARG(task) == param)
					flg = 1;
				break;
			case CRITERIA_FD:
				if (TASK_FD(task) == (intptr_t) param)
					flg = 1;
				break;
			case CRITERIA_ID:
			case CRITERIA_VAL:
				if (TASK_VAL(task) == (u_long) param)
					flg = 1;
				break;
			case CRITERIA_TS:
				if (!sched_timespeccmp(&TASK_TS(task), (struct timespec*) param, -))
					flg = 1;
				break;
			case CRITERIA_DATA:
				if (TASK_DATA(task) == param)
					flg = 1;
				break;
			case CRITERIA_DATLEN:
				if (TASK_DATLEN(task) == (size_t) param)
					flg = 1;
				break;
			default:
				sched_SetErr(EINVAL, "Invalid parameter criteria %d", criteria);
				flg = -1;
		}
		if (flg < 0)		/* error */
			break;
		/* cancel choosen task */
		if (flg > 0) {
			if (TASK_ROOT(task)->root_hooks.hook_exec.cancel)
				if (TASK_ROOT(task)->root_hooks.hook_exec.cancel(task, NULL)) {
					flg = -1;
					break;
				}
			/* custom hook */
			if (hook)
				if (hook(task, NULL)) {
					flg = -3;
					break;
				}

			TAILQ_REMOVE(queue, task, task_node);
			if (TASK_TYPE(task) != taskUNUSE)
				sched_unuseTask(task);

			flg ^= flg;	/* ok */
		}
	}
	SCHED_QUNLOCK(root, type);

	return flg;
}

/*
 * schedRun() - Scheduler *run loop*
 *
 * @root = root task
 * @killState = kill condition variable, if !=0 stop scheduler loop
 * return: -1 error or 0 ok
 */
int
schedRun(sched_root_task_t *root, volatile intptr_t * __restrict killState)
{
	sched_task_t *task;

	if (!root)
		return -1;

	if (root->root_hooks.hook_exec.run)
		if (root->root_hooks.hook_exec.run(root, NULL))
			return -1;

	if (killState) {
		if (root->root_hooks.hook_exec.condition)
			/* condition scheduler loop */
			while (root && root->root_hooks.hook_exec.fetch && 
					root->root_hooks.hook_exec.condition && 
					root->root_hooks.hook_exec.condition(root, (void*) killState)) {
				if ((task = root->root_hooks.hook_exec.fetch(root, NULL)))
					root->root_ret = schedCall(task);
			}
		else
			/* trigger scheduler loop */
			while (!*killState && root && root->root_hooks.hook_exec.fetch) {
				if ((task = root->root_hooks.hook_exec.fetch(root, NULL)))
					root->root_ret = schedCall(task);
			}
	} else
		/* infinite scheduler loop */
		while (root && root->root_hooks.hook_exec.fetch)
			if ((task = root->root_hooks.hook_exec.fetch(root, NULL)))
				root->root_ret = schedCall(task);

	return 0;
}

/*
 * schedPolling() - Polling timeout period if no timer task is present
 *
 * @root = root task
 * @ts = timeout polling period, if ==NULL INFINIT timeout
 * @tsold = old timeout polling if !=NULL
 * return: -1 error or 0 ok
 */
int
schedPolling(sched_root_task_t * __restrict root, struct timespec * __restrict ts, 
		struct timespec * __restrict tsold)
{
	if (!root)
		return -1;

	if (tsold)
		*tsold = root->root_poll;

	if (!ts)
		sched_timespecinf(&root->root_poll);
	else
		root->root_poll = *ts;

	return 0;
}

/*
 * schedTermCondition() - Activate hook for scheduler condition kill
 *
 * @root = root task
 * @condValue = condition value, kill schedRun() if condValue == killState
 * return: -1 error or 0 ok
 */
int
schedTermCondition(sched_root_task_t * __restrict root, intptr_t * __restrict condValue)
{
	if (!root && !condValue)
		return -1;

	*root->root_cond = *condValue;
	root->root_hooks.hook_exec.condition = sched_hook_condition;
	return 0;
}

/*
 * schedResumeby() - Resume suspended task
 *
 * @root = root task
 * @criteria = find task by criteria 
 * 	[CRITERIA_ANY|CRITERIA_ID|CRITERIA_VAL|CRITERIA_DATA]
 * @param = search parameter (sched_task_t *task| unsigned long id)
 * return: -1 error or 0 resumed ok
 */
int
schedResumeby(sched_root_task_t * __restrict root, u_char criteria, void *param)
{
	sched_task_t *task, *tmp;
	register int flg = 0;

	if (!root)
		return -1;

	SCHED_QLOCK(root, taskSUSPEND);
	TAILQ_FOREACH_SAFE(task, &root->root_suspend, task_node, tmp) {
		flg ^= flg;
		switch (criteria) {
			case CRITERIA_ANY:
				flg = 1;
				break;
			case CRITERIA_ID:
			case CRITERIA_VAL:
				if (TASK_VAL(task) == (u_long) param)
					flg = 1;
				break;
			case CRITERIA_DATA:
				if (TASK_ID(task) == (sched_task_t*) param)
					flg = 1;
				break;
			default:
				sched_SetErr(EINVAL, "Invalid parameter criteria %d", criteria);
				flg = -1;
		}
		if (flg < 0)
			break;
		/* resume choosen task */
		if (flg > 0) {
			if (root->root_hooks.hook_exec.resume)
				if (root->root_hooks.hook_exec.resume(task, NULL)) {
					flg = -1;
					break;
				}

			TAILQ_REMOVE(&root->root_suspend, task, task_node);

			task->task_type = taskREADY;
			insert_task_to(task, &root->root_ready);

			flg ^= flg;	/* ok */
		}
	}
	SCHED_QUNLOCK(root, taskSUSPEND);

	return flg;
}
